*******************************************************************************
*                         68008 Grundprogramm hardd                           *
*                         (C) 1990 Ralph Dombrowski                           *
*                             2007 Jens Mewes                                 *
*                                 Rev 7.00                                    *
*                                01.09.2007                                   *
*                            Harddisk-Routinen                                *
*******************************************************************************


* d1 = Befehlscode (SB = Basis-Befehlssatz / SE = Erweiterter Befehlssatz)
*  0 = Rezero Unit        (SB)
*  1 = Read               (SB)     2 = write                      (SB)
*  3 = Read Long          (SE)     4 = write Long                 (SE)
*  5 = Mode Select        (SB)     6 = Mode Sense                 (SB)
*  7 = Seek               (SB)     8 = Test Unit Ready            (SB)
*  9 = Park               (SB)    10 = Unpark                     (SB)
* 11 = Extended Read      (SE)    12 = Extended write             (SE)
* 13 = Read Buffer        (SE)    14 = write Buffer               (SE)
* 15 = Reserve            (SB)    16 = Release                    (SB)
* 17 = write + Verify     (SE)    18 = Verify                     (SE)
* 19 = Send Diagnostic    (SB)    20 = Extended Seek              (SE)
* 21 = Read Usage Counter (SB)    22 = Read Capacity              (SB)
* 23 = Rec. Diagnostic R. (SE)    24 = Inquiry                    (SB)
* 25 = Read Defect Data   (SB)    26 = Reassign Blocks            (SB)
* 27 = Request Sense      (SB)    28 = Format Unit                (SB)
* 29 = Eigene Befehle

* d2, d3 Je nach Befehl verschieden
* d4 = Laufwerksnummer
* a0.l = Adresse Daten, wenn verlangt
* a1.l = Bei Funktion 29 Adresse des Befehls

harddisk:
 cmp.b #1, scsi2ide(a5)         * Umleitung aktiv?
 beq idedisk                    * ja!
 btst.b #4,keydil(a5)           * Harddisk vorhanden ?
 beq hderr                      * Nein, dann Fehler
 movem.l d1-d5/a0-a3,-(a7)
 bsr.s hdcommand                * Befehlsauswertung aufrufen
 movem.l (a7)+,d1-d5/a0-a3
rts

hdbeftab:                       * Tabelle der Befehle
 dc.w hdgrup0-hdbeftab          * Auf Track 0
 dc.w hdgrup0b-hdbeftab         * Sektor lesen (d2.l/d3.b/a0.l)
 dc.w hdgrup0b-hdbeftab         * Sektor schreiben (d2.l/d3.b/a0.l)
 dc.w hdgrup0a-hdbeftab         * Sektor + ECC lesen (d2.l/a0.l)
 dc.w hdgrup0a-hdbeftab         * Sektor + ECC schreiben (d2.l/a0.l)
 dc.w hdbef5-hdbeftab           * Mode auswhlen (d2.b/a0.l)
 dc.w hdbef6-hdbeftab           * Parameter des Laufwerks lesen (d2.b/d3.b/a0.l)
 dc.w hdgrup0a-hdbeftab         * Sektor suchen (d2.l)
 dc.w hdgrup0-hdbeftab          * Laufwerk breit ?
 dc.w hdgrup0-hdbeftab          * Park
 dc.w hdbef10-hdbeftab          * Unpark
 dc.w hdgrup1b-hdbeftab         * Sektor lesen (d2.l/d3.w/a0.l)
 dc.w hdgrup1b-hdbeftab         * Sektor schreiben (d2.l/d3.w/a0.l)
 dc.w hdbef13-hdbeftab          * Buffer lesen (d2.w/a0.l)
 dc.w hdbef14-hdbeftab          * Buffer schreiben (d2.w/a0.l)
 dc.w hdbef15-hdbeftab          * Einheit reservieren (d2.w/d3.w/a0.l)
 dc.w hdbef16-hdbeftab          * Einheit freigeben (d2.w)
 dc.w hdgrup1b-hdbeftab         * Sektoren schreiben und prfen (d2.l/d3.w/a0.l)
 dc.w hdgrup1b-hdbeftab         * Sektor prfen (d2.l/d3.w)
 dc.w hdbef19-hdbeftab          * Diagnostic senden
 dc.w hdgrup1a-hdbeftab         * Sektor suchen (d2.l)
 dc.w hdgrup1-hdbeftab          * Zhler-Statistik lesen (a0.l)
 dc.w hdbef22-hdbeftab          * Gre der Platte lesen (d2.l/d3.b/a0.l)
 dc.w hdgrup0-hdbeftab          * Internen Test durchfhren
 dc.w hdbef24-hdbeftab          * Laufwerksnamen lesen (a0.l)
 dc.w hdbef25-hdbeftab          * Liste der Defekte lesen (d2.b/d3.w/a0.l)
 dc.w hdgrup0-hdbeftab          * Neue defekte Blcke schreiben (a0.l)
 dc.w hdbef27-hdbeftab          * Fehler lesen
 dc.w hdbef28-hdbeftab          * Formatieren (d2.b/d3.w/a0.l)

hdcomtab:                       * Tabelle der Kommandos
 dc.b $01,$08,$0A,$E5,$E6,$15,$1A,$0B,$00,$1B,$1B,$28,$2A,$3C,$3B,$16
 dc.b $17,$2E,$2F,$1D,$2B,$11,$25,$1C,$12,$37,$07,$03,$04
 ds 0

hdcommand:                      * Befehl aufbereiten und aufrufen
 cmp #29,d1
 beq.s hdcomm                   * Falls Befehl selber durchgefhrt werden soll
 bhi hderr                      * Wert ist zu gro
 lea hdtab(a5),a1               * Hier wird Befehl abgelegt
 moveq #0,d0                    * d0.l mu 0 sein, da als Langwort addiert wird
 move.b hdcomtab(pc,d1.w),d0    * Befehl
 add d1,d1
 move hdbeftab(pc,d1.w),d1
 jsr hdbeftab(pc,d1.w)          * Befehl aufbereiten
 clr.b (a1)                     * Ende ist immer gleich
 lea hdtab(a5),a1               * Adresse Befehl

hdcomm:                         * d4=Laufwerk / a0=Daten / a1=Adresse Command
 lea hddata.w,a2                * Adresse Daten-bertragungsregister
 lea hdstat.w,a3                * Adresse Status-Register
hdcom0:
 btst.b #1,(a3)                 * Warten, bis nicht mehr BUSY
 bne.s hdcom0
 move.b d4,(a2)                 * Laufwerks ID
 clr.b hdsel.w                  * Laufwerk aktivieren
hdcomin:                        * Einsprung fr Hardtest
 moveq #0,d0                    * Lschen fr Status, da Langwort gltig
*move.b (a3),d0                 * Hier Abfrage fr schnelle Erweiterung der
*and.b #%01100000,d0            * Harddisk (Hardware-Codierung)
*cmp.b #%00100000,d0            * Kennung fr spezielle Codierung
*beq.s hdcom8                   * Hier besonders schnelle Routine
 moveq #0,d5                    * Request-Bit
 moveq #3,d4                    * Data/Command
 moveq #4,d3                    * Input/Output
 moveq #2,d2                    * Message-Bit
hdcom1:
 btst.b #1,(a3)                 * Warten, bis ausgewhlt
 beq.s hdcom1
hdcom2:                         * Ohne Erweiterung (normale Geschwindigkeit)
 move.b (a3),d1
 btst d5,d1
 beq.s hdcom2                   * Schleife, bis Request vorhanden
 btst d4,d1                     * Auswahl Daten oder Command
 bne.s hdcom4
 btst d3,d1                     * Auswahl Input oder Output
 beq.s hdcom3
 move.b (a2),(a0)+              * Data in
bra.s hdcom2
hdcom3:
 move.b (a0)+,(a2)              * Data out
bra.s hdcom2

hdcom4:                         * Command
 btst d2,d1                     * Message oder Command/Status
 bne.s hdcom6
 btst d3,d1                     * Input oder Output
 bne.s hdcom5
 move.b (a1)+,(a2)              * Command
bra.s hdcom2
hdcom5:                         * Status
 move.b (a2),d0                 * Merken fr Fehlerabfrage
bra.s hdcom2

hdcom6:                         * Message
 btst d3,d1                     * Input oder Output
 bne.s hdcom7
 move.b #$08,(a2)               * Message out (NO OPERATION)
bra.s hdcom2

hdcom8:                         * Hier fr Hardware-Codierung
*btst.b #1,(a3)                 * Warten, bis ausgewhlt
*beq.s hdcom8
*moveq #0,d1                    * Langwort soll gltig sein
*lea hdcode.w,a3                * Diese Adresse ist fr die Erweiterung
hdcom9:
*move.b (a3),d1                 * Relative Sprungadresse holen
*jmp hdcom9(pc,d1.l)            * hinspringen

* Von hier an darf nichts verndert werden *

*move.b (a2),(a0)+              * Data in
*bra.s hdcom9
*move.b (a0)+,(a2)              * Data out
*bra.s hdcom9
*move.b (a1)+,(a2)              * Command
*bra.s hdcom9
*move.b (a2),d0                 * Status
*bra.s hdcom9
*move.b #$08,(a2)               * Message out
*bra.s hdcom9
hdcom7:
 tst.b (a2)                     * Message in (Ende)
bra carres                      * Immer Carry=0 (Fehlercode in d0)

* Bis hierhin darf nichts gendert werden *

hdbef5:                         * Mode Auswahl
 moveq #0,d3                    * d2.b = Lnge der Parameter Liste/a0.l = Daten
 move.b d2,d3                   * Keine Sekorenauswahl
 moveq #0,d2
bra.s hdgrup0b

hdbef6:                         * Laufwerksdaten lesen
 and.l #$ff,d2                  * d2.b = PC + Page-Code
 lsl.l #8,d2                    * d3.b = Lnge der Daten
 and.w #$ff,d3
bra.s hdgrup0b

hdbef10:                        * Unpark
 moveq #0,d2                    * Meldung erst, wenn wirklich zurck
 moveq #1,d3                    * Auf Track 0
bra.s hdgrup0b

hdbef13:                        * Buffer lesen
hdbef14:                        * Buffer schreiben
 move d2,d3                     * d2.w = Maximale Lnge
 moveq #0,d2
bra.s hdgrup0b

hdbef16:                        * Einheit freigeben
 moveq #0,d3                    * Keine Lngenangabe
hdbef15:                        * Einheit reservieren
 lsl.l #8,d2                    * d2.b = Identifikation
bra.s hdgrup0b                  * d3.w = Lnge der Daten

hdbef19:                        * Diagnostic senden
 moveq #4,d2
 swap d2                        * $00040000 -> d2
 moveq #0,d3
bra.s hdgrup0b

hdbef22:                        * Gre der Platte lesen
 and.w #1,d3                    * d2.l = Sektornummer / d3.b = PMI
bra.s hdgrup1b                  * a0.l = Adresse Ergebnis

hdbef24:                        * Laufwerksdaten lesen
 moveq #0,d2
 moveq #35,d3                   * Alle Daten lesen
bra.s hdgrup0b

hdbef25:                        * Liste der Defekte lesen
 and.l #3,d2                    * d2.b = Auswahl
 ror.l #5,d2                    * d3.w = Maximale Lnge
bra.s hdgrup1b                  * a0.l = Adresse Daten

hdbef27:                        * Fehler lesen
 moveq #0,d2                    * Keine Sektornummer
 moveq #27,d3                   * Maximal 27 Bytes bertragen
bra.s hdgrup0b

hdbef28:                        * Formatieren
 move.b d0,(a1)+                * d2.b = Format der Defect-List
 and #$1f,d2                    * a0.l = Defect List
 move.b d2,(a1)+                * d3.w = Interleave
bra.s hdgrup1c                  * Rest

hdgrup0:
 moveq #0,d2                    * Keine Sektornummer erforderlich
hdgrup0a:
 moveq #0,d3                    * Keine Anzahl erforderlich
hdgrup0b:
 and.l #$001fffff,d2            * Sektornummer o..
 ror.l #8,d0
 add.l d0,d2                    * Befehlscode dazu
 move.l d2,(a1)+                * Ablage
 move.b d3,(a1)+                * Anzahl Blcke oder Daten
rts

hdgrup1:
 moveq #0,d2                    * Keine Sektoranzahl erforderlich
hdgrup1a:
 moveq #0,d3                    * Keine Anzahl erforderlich
hdgrup1b:
 move.b d0,(a1)+                * Befehlscode
 clr.b (a1)+                    * LUN und Reserve
 move.l d2,(a1)+                * Sektornummer o..
hdgrup1c:
 clr.b (a1)+                    * Reserviert
 rol #8,d3
 move.b d3,(a1)+
 rol #8,d3
 move.b d3,(a1)+                * Anzahl Blcke oder Daten
rts

hardtest:                       * Test, ob das Laufwerk vorhanden ist
 cmp.b #1, scsi2ide(a5)         * Umleitung aktiv?
 beq idetest                    * ja!
 btst.b #4,keydil(a5)           * Harddisk vorhanden ?
 beq hderr                      * Nein, dann Fehler
hardt0:
 btst.b #1,hdstat.w             * d4.b = Laufwerksnummer
 bne.s hardt0                   * Warten, bis nicht mehr BUSY
 move.b d4,hddata.w             * Laufwerks ID
 clr.b hdsel.w                  * Laufwerk aktivieren
 move #2000*cpu,d0              * Schleifendurchlufe sind CPU-abhngig
hdtest1:
 btst.b #1,hdstat.w             * Warten, bis ausgewhlt
 bne.s hdtest2                  * OK, Laufwerk vorhanden
 dbra d0,hdtest1
bra hderr                       * Laufwerk nicht vorhanden (antwortet nicht)
hdtest2:
 movem.l d1-d5/a0-a3,-(a7)
 lea hdtab+4(a5),a1             * TEST UNIT READY
 clr.w (a1)
 clr.l -(a1)
 lea hddata.w,a2                * Adresse Daten-bertragungsregister
 lea hdstat.w,a3                * Adresse Status-Register
 bsr hdcomin
 movem.l (a7)+,d1-d5/a0-a3
rts

hderr:
 moveq #-1,d0                   * Error Code
bra carset                      * Carry auch setzen


idetest:                        * Test, ob Laufwerk vorhanden
                                * d4 enthaelt Laufwerk
 btst.b #5,keydil(a5)           * IDE vorhanden ?
 beq.s hderr                    * Nein, dann Fehler
 tst.b d4
 beq.s hderr                    * Kein LW angegeben
 and.b #$0f, d4
 movem.l d1-d4/a1/a6, -(a7)
 cmp.b #1, d4
 bne.s idt01
 lea idemgeo(a5), a6            * Variablenspeicher LW1
 move #$a0, d4                  * d4 jetzt Master-Flag
 bra.s idt02
idt01:                          * Slave-Laufwerk
 cmp.b #2, d4
 bne.s idterr                   * kein gueltiges Laufwerk
 lea idesgeo(a5), a6            * Variablenspeicher Slave
 move #$b0, d4                  * d4 jetzt Slave-Flag
idt02:
 bsr ideid                      * LW Identifizieren
 tst d0
 bmi.s idterr                   * Fehler bei Identifizieren
 bsr ideinit                    * LW initialisieren
 tst d0
 bmi.s idterr                   * Fehler beim Initialisieren
 movem.l (a7)+, d1-d4/a1/a6
 movea.l a6, a0
bra carres

idterr:
 movem.l (a7)+, d1-d4/a1/a6
 bra.s hderr

ideid:                          * Identifiziert ein Laufwerk
 move.b d4, idesdh.w            * Master / Slave
idlp01:
 clr numcyl(a6)                 * LW-Parameter lschen
 clr.b numhead(a6)
 clr.b numsec(a6)
 bsr idewr
 tst d0
 bmi idlp20
 move.b #cmdident, idecmd.w     * LW Identifizierung
 bsr idewd                      * Daten bereit?
 tst d0
 bmi idlp20                     * nein, dann Fehler
 lea idebuff(a5),a0             * Puffer fr Transfer
 lea idedat.w, a1
 move.l #512-1, d3
idlp04:
 move.b (a1), (a0)+             * Ident-Daten einlesen
 dbra d3, idlp04
 bsr idewr                      * LW fertig?
 tst d0
 bmi idlp20                     * nein, dann Fehler
 lea idebuff(a5),a0             * Puffer fr Transfer
 clr.l d0
 move.b 3(a0), d0               * Zylinder High-Byte
 lsl #8, d0
 move.b 2(a0), d0               * Zylinder Low-Byte
 move d0, numcyl(a6)            * speichern
 clr.l d0
 move.b 6(a0), numhead(a6)      * Anzahl der Kpfe
 move.b 12(a0), numsec(a6)      * Anzahl der Sektoren/Spur
 movea.l a6, a1
 adda.l #idename, a1            * a1 auf LW-Name-Speicher
 adda.l #54, a0                 * A0 auf Namensquelle
 move #12-1, d3                 * nur die ersten 24 Zeichen
idlp10:
 move.b 1(a0), (a1)+            * Transfer mit Byteswap
 move.b 0(a0), (a1)+
 addq.l #2, a0
 dbra d3, idlp10
 move #$0, (a1)+                * noch ne 0 zum Abschluss
 bra.b idex
idlp20:
 move #-1, d0
idex:
 rts

ideinit:                        * Initialisiert das Laufwerk
 move.b #6, idedor.w
 move.l #255, d3
ideilp01:
 dbra d3, ideilp01              * ein bisschen warten
 move.b #2, idedor.w
 bsr idewr                      * LW bereit?
 tst d0
 beq.b ideilp03                 * ja
 bra ideiex
ideilp03:
 move.b numsec(a6), idescnt.w   * Anzahl Sektoren
 move.b numcyl+1(a6), ideclo.w  * Zylinder Low-Byte
 move.b numcyl(a6), idechi.w    * Zylinder High-Byte
 move.b numhead(a6), d0         * Anzahl Kpfe
 subq.b #1, d0                  * -1
 or.b d4, d0                    * mit LW Kennung verodert
 move.b d0, idesdh.w            * und ausgeben
 move.b #cmdinit, idecmd.w      * nun weis das LW was es ist ;)
 clr d0
ideiex:
 rts
 

idebeftab:                      * Tabelle der Befehle
 dc.w ideok-idebeftab           * Auf Track 0
 dc.w idebef1-idebeftab         * Sektor lesen (d2.l/d3.b/a0.l)
 dc.w idebef2-idebeftab         * Sektor schreiben (d2.l/d3.b/a0.l)
 dc.w idenok-idebeftab          * Sektor + ECC lesen (d2.l/a0.l)
 dc.w idenok-idebeftab          * Sektor + ECC schreiben (d2.l/a0.l)
 dc.w ideok-idebeftab           * Mode auswhlen (d2.b/a0.l)
 dc.w ideok-idebeftab           * Parameter des Laufwerks lesen (d2.b/d3.b/a0.l)
 dc.w ideok-idebeftab           * Sektor suchen (d2.l)
 dc.w idebef8-idebeftab         * Laufwerk breit ?
 dc.w ideok-idebeftab           * Park
 dc.w ideok-idebeftab           * Unpark
 dc.w idenok-idebeftab          * Sektor lesen (d2.l/d3.w/a0.l)
 dc.w idenok-idebeftab          * Sektor schreiben (d2.l/d3.w/a0.l)
 dc.w idenok-idebeftab          * Buffer lesen (d2.w/a0.l)
 dc.w idenok-idebeftab          * Buffer schreiben (d2.w/a0.l)
 dc.w ideok-idebeftab           * Einheit reservieren (d2.w/d3.w/a0.l)
 dc.w ideok-idebeftab           * Einheit freigeben (d2.w)
 dc.w idenok-idebeftab          * Sektoren schreiben und prfen (d2.l/d3.w/a0.l)
 dc.w idenok-idebeftab          * Sektor prfen (d2.l/d3.w)
 dc.w ideok-idebeftab           * Diagnostic senden
 dc.w idenok-idebeftab          * Sektor suchen (d2.l)
 dc.w ideok-idebeftab           * Zhler-Statistik lesen (a0.l)
 dc.w idebef22-idebeftab        * Gre der Platte lesen (d2.l/d3.b/a0.l)
 dc.w idenok-idebeftab          * Internen Test durchfhren
 dc.w idebef24-idebeftab        * Laufwerksnamen lesen (a0.l)
 dc.w ideok-idebeftab           * Liste der Defekte lesen (d2.b/d3.w/a0.l)
 dc.w ideok-idebeftab           * Neue defekte Blcke schreiben (a0.l)
 dc.w ideok-idebeftab           * Fehler lesen
 dc.w ideok-idebeftab           * Formatieren (d2.b/d3.w/a0.l)
 
idedisk:                        * IDE IO-Routine
 btst.b #5,keydil(a5)           * IDE vorhanden ?
 beq hderr                      * Nein, dann Fehler
 movem.l d1-d5/a0-a3/a6, -(a7)
 bsr.s idecomm                  * Hauptroutine aufrufen
 movem.l (a7)+, d1-d5/a0-a3/a6
rts

idecomm:
 cmp #29, d1
 beq ideok                      * keine Eigenen Befehle
 bhi hderr                      * Wert zu gross
 cmp #1, d4                     * Master?
 bne.s idec1                    * n
 lea idemgeo(a5), a6
 bra.s idec2
idec1:
 cmp #2, d4                     * Slave?
 bne.s idec2                    * n
 lea idesgeo(a5), a6
idec2:
 add d1, d1                     * mal 2 da Wort
 move idebeftab(pc,d1.w), d1    * Sprungwert laden
 jsr idebeftab(pc,d1.w)
 bra carres

ideok:                          * liefert nur ein OK zurck
 clr d0
 bra carres

idenok:                         * liefert einen Fehler zurck
 moveq #-1, d0
 rts

idebef1:                        * Sektoren lesen
 cmp #1, d4                     * Master LW?
 bne.s idb1a                    * nein
 move #$a0, d4
 bra.s idb1b
idb1a:
 cmp #2, d4                     * Slave LW?
 bne idb1err                    * nein, dann Fehler
 move #$b0, d4
idb1b:
 move.l d2, d0                  * Startsektor
 move d3, d1                    * Anzahl
 tst.b scsi2ide(a5)             * Umleitung aktiv?
 beq.s idb1c                    * nein
 asl.l #1, d0                   * Startsektor *2
 asl #1, d1                     * Anzahl * 2
 tst d1                         * Null?
 bne.s idb1e                    * NEIN! dann keine 512 Sektoren
 move.l #256, d2
 bra.s idb1f
idb1e:
 cmp #256, d1                   * < 256?
 blt.s idb1c                    * ja
idb1d:
 clr.l d2
 lsr #1, d1                     * Anzahl /2, da 2 Aufrufe
 move d1, d2                    * sichern
idb1f:
 move.l d0, d3                  * sichern
 bsr iderdsek                   * 1. Lesen, bei mehr als 256 Sektoren
 tst d0
 bne.s idb1err
 move.l d3, d0                  * Startsektor zurck
 add.l d2, d0                   * um Anzahl Sektoren erhhen
 mulu #512, d2                  * Anzahl * Grsse
 adda.l d2, a0                  * Puffer erhhen
idb1c:
 bsr iderdsek                   * Lesen, zum 2. bei Sektoren > 256
 tst d0
 beq.s idb1ex
idb1err:
 move #-1, d0
idb1ex:
 rts

idebef2:                        * Sektoren schreiben
 cmp #1, d4                     * Master LW?
 bne.s idb2a                    * nein
 move #$a0, d4
 bra.s idb2b
idb2a:
 cmp #2, d4                     * Slave LW?
 bne idb2err                    * nein, dann Fehler
 move #$b0, d4
idb2b:
 move.l d2, d0                  * Startsektor
 move d3, d1                    * Anzahl
 tst.b scsi2ide(a5)             * Umleitung aktiv?
 beq.s idb2c                    * nein
 asl.l #1, d0                   * Startsektor *2
 asl #1, d1                     * Anzahl * 2
 tst d1                         * Null?
 bne.s idb2e                    * NEIN! dann keine 512 Sektoren
 move.l #256, d2
 bra.s idb2f
idb2e: 
 cmp #256, d1                   * < 256?
 blt.s idb2c                    * ja
idb2d:
 clr.l d2
 lsr #1, d1                     * Anzahl /2, da 2 Aufrufe
 move d1, d2                    * sichern
 addq #1, d2                    * Anzahl + 1 (1...n)
idb2f:
 move.l d0, d3                  * sichern
 bsr idewrsek                   * 1. Schreiben, bei mehr als 256 Sektoren
 tst d0
 bne.s idb2err
 move.l d3, d0                  * Startsektor zurck
 add.l d2, d0                   * um Anzahl Sektoren erhhen
 mulu #512, d2                  * Anzahl * Grsse
 adda.l d2, a0                  * Puffer erhhen
idb2c:
 bsr idewrsek                   * Schreiben, zum 2. bei Sektoren > 256
 tst d0
 beq.s idb2ex
idb2err:
 move #-1, d0
idb2ex:
 rts

idebef8:                        * Laufwerk bereit?
 btst.b #7, idecmd.w            * Busy-Flag abfragen
 beq.s idb8a
 move.l #4, d0                  * und entsprechend SCSI setzen
 bra.s idb8ex
idb8a:
 clr.l d0                       * oder lschen
idb8ex:
 rts

idebef22:                       * Kapazitt lesen
 clr.l d0
 move.b numsec(a6), d0
 move.l #512, d1                * 512 Byte/Sektor
 cmp #1, d3                     * Sektoren pro Spur
 bne.s idb22a                   * nein
 tst.b scsi2ide(a5)
 beq.s idb22ex                  * Umleitung nicht aktiv
 lsr.l #1, d0                   * /2 da 1024 Byte pro Sektor
 asl.l #1, d1                   * auf 1024 Bytes setzen
 bra.s idb22ex
idb22a:
 move.b numhead(a6), d3
 mulu d3, d0                    * Kpfe * Sektoren
 move numcyl(a6), d3
 mulu d3, d0                    * Kpfe * Sektoren * Spuren
 tst.b scsi2ide(a5)
 beq.s idb22ex                  * Umleitung nicht aktiv
 lsr.l #1, d0                   * /2 da 1024 Byte pro Sektor
 asl.l #1, d1                   * auf 1024 Bytes setzen
idb22ex:
 move.l d0, 0(a0)
 move.l d1, 4(a0)
 clr.l d0
 rts

idebef24:                       * LW Name lesen
 move #36-1, d3                 * 36 Byte Buffer
 movea.l a0, a1
idb24a:
 clr.b (a1)+                    * lschen
 dbra d3, idb24a
 move.b #1, 3(a0)               * ??? aus SCSI Bescheibung bernommen
 move.b #$3d, 4(a0)             * ??? aus SCSI Bescheibung bernommen
 move #24-1, d3                 * 24 Byte bertragen
 movea.l a6, a1                 * ide_geo
 adda.l #idename, a1
 adda.l #8, a0
idb24b:
 move.b (a1)+, (a0)+            * Name kopieren
 dbra d3, idb24b
 rts
 

iderdsek:                * Sektor(en) lesen D0=Sektornr., D1=Anzahl, A0=Puffer
 movem.l d1-d5/a1-a2, -(a7)
 move.l d0, d2                          * sichern
 move.l d1, d5                          * sichern
 bsr idewr                              * LW bereit?
 tst d0
 beq.b rdlp01                           * ja
 bra rderr                              * sonst Fehler
rdlp01:
 move.b d5, idescnt.w                   * d5 Sektor(en)
 move.l d2, d0                          * Sektor zurck
 bsr lba2chs                            * nach CHS umrechnen
 move.b d2, idesnum.w                   * Start-Sektor
 move.b d3, ideclo.w                    * Start Spur Low-Byte
 lsr #8, d3
 move.b d3, idechi.w                    * Start-Spur High-Byte
 or.b d4, d1                            * Anzahl Kpfe mit LW verodert
rdlp10:
 move.b d1, idesdh.w                    * und bergeben
 tst d5                                 * Sektorenanzahl=0?
 bne.s rdlp11                           * nein
 move #256, d5                          * sonst = 256
rdlp11:
 move #10-1, d2                         * 10 Versuche
 movea.l a0, a2                         * retten
 subq #1, d5                            * als Zhler
 move d5, d4                            * retten
 lea idedat.w, a1                       * Transferadresse nach a1
rdlp12:
 move d4, d5                            * wiederherstellen
 movea.l a2, a0                         * dito
 move.b #cmdrd, idecmd.w                * Lese-Befehl
rdlp12a:
 move #512-1, d3                        * Anzahl Bytes -1
 bsr idewd                              * Daten bereit?
 tst d0
 bne rderr                              * nein, dann Fehler
rdlp13:
 move.b (a1), (a0)+                     * lesen
 dbra d3, rdlp13                        * nchstes Byte
 dbra d5, rdlp12a                       * nchsten Sektor
 bsr idewr                              * LW fertig?
 tst d0
 bne rderr                              * nein, dann Fehler
 move.b idecmd.w, d0
 and.b #%10001001, d0                   * irgend welche weiteren Fehler?
 beq.b rdlp20                           * n, dann fertig
 dbra d2, rdlp12                        * sonst nochmal
rderr:
 moveq #-1, d0
 bra.b rdex
rdlp20:
 clr d0
rdex:
 movem.l (a7)+, d1-d5/a1-a2
 rts


idewrsek:            * Sektor(en) schreiben D0=Sektornr., d1=Anzahl, A0=Puffer
 movem.l d1-d5/a1-a2, -(a7)
 move.l d0, d2                          * sichern
 move.l d1, d5
 bsr idewr
 tst d0
 beq.b wrlp01
 bra wrerr
wrlp01:
 move.b d5, idescnt.w                   * d5 Sektor(en)
 move.l d2, d0                          * Sektor zurck
 bsr lba2chs                            * in CHS umrechnen
 move.b d2, idesnum.w
 move.b d3, ideclo.w
 lsr #8, d3
 move.b d3, idechi.w
 or.b d4, d1
 move.b d1, idesdh.w
 tst d5
 bne.s wrlp11
 move #256, d5
wrlp11:
 move #10-1, d2                         * 10 Versuche
 movea.l a0, a2                         * retten
 subq #1, d5                            * -1 als Zhler
 move d5, d4                            * retten
 lea idedat.w, a1
wrlp12:
 move d4, d5                            * wiederherstellen
 movea.l a2, a0
 move.b #cmdwr, idecmd.w
wrlp12a:
 move #512-1,d3                         * Anzahl Bytes -1
 bsr idewd
 tst d0
 bne wrerr
wrlp13:
 move.b (a0)+, (a1)                     * schreiben
 dbra d3, wrlp13                        * nchstes Byte
 dbra d5, wrlp12a                       * nchster Sektor
 bsr idewr
 tst d0
 bne wrerr
 move.b idecmd.w, d0
 and.b #%10001001, d0                   * irgend welche Fehler?
 beq.b wrlp20                           * n, fertig
 dbra d2, wrlp12                        * sonst noch'n Versuch
wrerr:
 moveq #-1, d0
 bra.b wrex
wrlp20:
 clr d0
wrex:
 movem.l (a7)+, d1-d5/a1-a2
 rts


lba2chs:                        * Rechenet LBA (d0.l) in Head (d1.b),
                                * Sektor (d2.b) und Zylinder (d3.w) um
                                * a6 = Master/Slave-Buffer
 clr.l d1
 move.b numsec(a6), d1
 divu d1, d0                            * log. Sektor / Sektoren pro Spur
 swap d0                                * nur der Rest
 move.b d0, d2                          * Sektor
 addq.b #1, d2                          * +1
 clr d0
 swap d0                                * Divisionsergebnis
 move.b numhead(a6), d1
 divu d1, d0                            * durch Anzahl der Kpfe
 move d0, d3                            * Zylinder
 swap d0                                * Kpfe
 move.b d0, d1
 rts

idewr:                                  * Warten bis Laufwerk ready
 move.l #idedel*cpu, d0                 * Delaywert laden
iwr01:
 subq.l #1, d0
 bmi iwr02                              * Abbruch
 btst.b #7, idecmd.w                    * Busy?
 bne.b iwr01                            * ja!
 clr d0                                 * ist ready
 bra.b iwrex
iwr02:
 move #-1, d0                           * ist NICHT ready
iwrex:
 rts

idewd:                                  * Warten bis LW bereit fr Daten
 move.l #idedel*cpu, d0                 * Delaywert laden
iwd01:
 subq.l #1, d0
 bmi iwd02                              * Abbruch
 btst.b #3, idecmd.w                    * bereit fr Datentransfer?
 beq.b iwd01                            * ja
 clr d0                                 * Daten bereit
 bra.b iwdex
iwd02:
 move #-1, d0                           * Daten NICHT bereit
iwdex:
 rts


numcyl     equ 0
numhead    equ 2
numsec     equ 3
nkcmode    equ 4
idename    equ 6

idedel     equ 40000

cmdrd      equ $20
cmdwr      equ $30
cmdinit    equ $91
cmdident   equ $ec


sets2i:                         * SCSI nach IDE Umleitung setzen
 btst.b #5,keydil(a5)           * IDE vorhanden ?
 beq.s sets2ier                 * Nein, dann Fehler
 move.b d0, scsi2ide(a5)
 clr d0
 rts
sets2ier:
 move #-1, d0
 rts
 
gets2i:                         * SCSI nach IDE Umleitung laden
 move.b scsi2ide(a5), d0
 rts